--thanks to http://www.mit.edu/~tarvizo/nes-controller.html

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;

entity nes_ctlr is
	port (
	    avs_s1_clk        : in  std_logic;
	    avs_s1_reset_n    : in  std_logic;
    	avs_s1_read       : in  std_logic;
	    avs_s1_write      : in  std_logic;
   		avs_s1_chipselect : in  std_logic;
	    avs_s1_address    : in  std_logic;
    	avs_s1_readdata   : out std_logic_vector(7 downto 0);
	    avs_s1_writedata  : in  std_logic_vector(7 downto 0);

		latch1: out std_logic;
		latch2: out std_logic;

		pulse1: out std_logic;
		pulse2: out std_logic;

		data1: in std_logic; --data
		data2: in std_logic --data

		);
end nes_ctlr;

architecture fsm of nes_ctlr is
	constant w : integer range 0 to 17 := 0;
    constant w0 : integer range 0 to 17 := 1;
    constant b0 : integer range 0 to 17 := 2;
    constant w1 : integer range 0 to 17 := 3;
    constant b1 : integer range 0 to 17 := 4;
    constant w2 : integer range 0 to 17 := 5;
    constant b2 : integer range 0 to 17 := 6;
    constant w3 : integer range 0 to 17 := 7;
    constant b3 : integer range 0 to 17 := 8;
    constant w4 : integer range 0 to 17 := 9;
    constant b4 : integer range 0 to 17 := 10;
    constant w5 : integer range 0 to 17 := 11;
    constant b5 : integer range 0 to 17 := 12;
    constant w6 : integer range 0 to 17 := 13;
    constant b6 : integer range 0 to 17 := 14;
    constant w7 : integer range 0 to 17 := 15;
    constant b7 : integer range 0 to 17 := 16;
    constant w8 : integer range 0 to 17 := 17;

	signal clk_60Hz: std_logic;
	signal clk_41KHz: std_logic;
	signal clk_83KHz: std_logic;
	signal counter_60Hz : integer range 0 to 8500000 := 0;
	signal counter_41KHz : integer range 0 to 300 := 0;
	signal counter_83KHz : integer range 0 to 600 := 0;
    signal p_s, n_s : integer range 0 to 17 := 0;
	signal player1: std_logic_vector (7 downto 0); --data
	signal player2: std_logic_vector (7 downto 0); --data
	signal latch_counter: integer range 0 to 300 := 0;
	signal pulse_counter: integer range 0 to 5400 := 0;
	signal latch_sig: std_logic;
	signal pulse_sig: std_logic;

begin

	Clock_60Hz:process(avs_s1_clk)
	begin
		if rising_edge(avs_s1_clk) then
			if (counter_60Hz = 416667) then
				if (clk_60Hz = '1') then
					clk_60Hz <= '0';
					counter_60Hz <= 0;
				else
					clk_60Hz <= '1';
					counter_60Hz <= 0;
				end if;
			else
				counter_60Hz <= counter_60Hz + 1;
			end if;
		end if;
	end process Clock_60Hz;

	Clock_41KHz:process(avs_s1_clk)
	begin
		if rising_edge(avs_s1_clk) then
			if (counter_41KHz = 150) then
				if (clk_41KHz = '1') then
					clk_41KHz <= '0';
					counter_41KHz <= 0;
				else
					clk_41KHz <= '1';
					counter_41KHz <= 0;
				end if;
			else
				counter_41KHz <= counter_41KHz + 1;
			end if;
		end if;
	end process Clock_41KHz;

	Clock_83KHz:process(avs_s1_clk)
	begin
		if rising_edge(avs_s1_clk) then
			if (counter_83KHz = 300) then
				if (clk_83KHz = '1') then
					clk_83KHz <= '0';
					counter_83KHz <= 0;
				else
					clk_83KHz <= '1';
					counter_83KHz <= 0;
				end if;
			else
				counter_83KHz <= counter_83KHz + 1;
			end if;
		end if;
	end process Clock_83KHz;

	Latch_Control:process (avs_s1_clk, clk_60Hz)
	begin
		if rising_edge (clk_60Hz) then
			latch1 <= '1';
			latch2 <= '1';
			latch_sig <= '1';
		elsif latch_counter < 300 then
			latch_counter <= latch_counter + 1;
		else 
			latch1 <= '0';
			latch2 <= '0';
			latch_sig <= '0';
			latch_counter <= 0;
		end if;
	end process Latch_Control;

	Pulse_Control:process (avs_s1_clk, clk_60Hz)
	begin
		if rising_edge (clk_60Hz) then
			pulse_counter <= 0;
		elsif pulse_counter < 600 then
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 900 then --1
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 1200 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 1500 then --2
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 1800 then 
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 2100 then --3
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 2400 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 2700 then --4
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 3000 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 3300 then --5
			pulse1 <= '1';
			pulse2 <= '1';			
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 3600 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 3900 then --6
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 4200 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 4500 then --7
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 4800 then
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';
			pulse_counter <= pulse_counter + 1;
		elsif pulse_counter < 5100 then --8
			pulse1 <= '1';
			pulse2 <= '1';
			pulse_sig <= '1';
			pulse_counter <= pulse_counter + 1;
		else --9
			pulse1 <= '0';
			pulse2 <= '0';
			pulse_sig <= '0';

		end if;
	end process Pulse_Control;

	fsm_process:process(p_s, latch_sig, pulse_sig)
	begin
        case p_s is
            when w =>
                if latch_sig = '1' then
                    n_s <= w0;
                else
                    n_s <= w;
                end if;
            when w0 =>
                if latch_sig = '0' then
                    n_s <= b0;
                    player1(0) <= not data1;
                    player2(0) <= not data2;
                else
                    n_s <= w0;
                end if;
            when b0 =>
                if pulse_sig = '1' then
                    n_s <= w1;
                else
                    n_s <= b0;
                end if;
            when w1 =>
                if pulse_sig = '0' then
                    n_s <= b1;
                    player1(1) <= not data1;
                    player2(1) <= not data2;
                else
                    n_s <= w1;
                end if;
            when b1 =>
                if pulse_sig = '1' then
                    n_s <= w2;
                else
                    n_s <= b1;
                end if;
            when w2 =>
                if pulse_sig = '0' then
                    n_s <= b2;
                    player1(2) <= not data1;
                    player2(2) <= not data2;
                else
                    n_s <= w2;
                end if;
            when b2 =>
                if pulse_sig = '1' then
                    n_s <= w3;
                else
                    n_s <= b2;
                end if;
            when w3 =>
                if pulse_sig = '0' then
                    n_s <= b3;
                    player1(3) <= not data1;
                    player2(3) <= not data2;
                else
                    n_s <= w3;
                end if;
            when b3 =>
                if pulse_sig = '1' then
                    n_s <= w4;
                else
                    n_s <= b3;
                end if;
            when w4 =>
                if pulse_sig = '0' then
                    n_s <= b4;
                    player1(4) <= not data1;
                    player2(4) <= not data2;
                else
                    n_s <= w4;
                end if;
            when b4 =>
                if pulse_sig = '1' then
                    n_s <= w5;
                else
                    n_s <= b4;
                end if;
            when w5 =>
                if pulse_sig = '0' then
                    n_s <= b5;
                    player1(5) <= not data1;
                    player2(5) <= not data2;
                else
                    n_s <= w5;
                end if;
            when b5 =>
                if pulse_sig = '1' then
                    n_s <= w6;
                else
                    n_s <= b5;
                end if;
            when w6 =>
                if pulse_sig = '0' then
                    n_s <= b6;
                    player1(6) <= not data1;
                    player2(6) <= not data2;
                else
                    n_s <= w6;
                end if;
            when b6 =>
                if pulse_sig = '1' then
                    n_s <= w7;
                else
                    n_s <= b6;
                end if;
            when w7 =>
                if pulse_sig = '0' then
                    n_s <= b7;
                    player1(7) <= not data1;
                    player2(7) <= not data2;
                else
                    n_s <= w7;
                end if;
            when b7 =>
                if pulse_sig = '1' then
                    n_s <= w8;
                else
                    n_s <= b7;
                end if;
            when w8 =>
                if pulse_sig = '0' then
                    n_s <= w;
                else
                    n_s <= w8;
                end if;
            when others =>
                if latch_sig = '1' then
                    n_s <= w0;
                end if;
        end case;
	end process fsm_process;
	
	state_transition:process (clk_60Hz)
	begin
		if rising_edge(clk_60Hz) then
			p_s <= n_s;
		end if;
	end process state_transition;
	
	process (clk_60Hz)
	begin
		if avs_s1_reset_n = '0' then
        	avs_s1_readdata <= (others => '0');
      	else
        	if avs_s1_chipselect = '1' then
          		if avs_s1_read = '1' then
					if avs_s1_address = '0' then
            			avs_s1_readdata <= player1;	
					else
						avs_s1_readdata <= player2;
					end if;
          		end if;
        	end if;
      	end if;
  	end process;
	
end fsm;